@@ -110,16 +110,14 @@ call_r_function <- function(fn, args, named_args) {
       python.builtin.BaseException = function(e) {
         # check if rethrowing an exception that we've already seen
         # and if so, make sure the r_trace attr is still present
-        r_trace <- as_r_value(py_get_attr(e, "r_trace", TRUE))
+        r_trace <- as_r_value(py_get_attr(e, "trace", TRUE))
         if(is.null(r_trace)) {
           r_trace <- get_r_trace(maybe_use_cached = TRUE, trim_tail = 2)
-          py_set_attr(e, "r_trace", py_capsule(r_trace))
+          py_set_attr(e, "trace", py_capsule(r_trace))
         }
 
-        if(!py_has_attr(e, "r_call")) {
-          if(is.null(r_trace))
-            r_trace <- get_r_trace(maybe_use_cached = TRUE, trim_tail = 2)
-          py_set_attr(e, "r_call", py_capsule(r_trace$full_call[[nrow(r_trace)]]))
+        if(!py_has_attr(e, "call")) {
+          py_set_attr(e, "call", py_capsule(r_trace$full_call[[nrow(r_trace)]]))
         }
 
         invokeRestart("raise_py_exception", e)
@@ -131,45 +129,44 @@ call_r_function <- function(fn, args, named_args) {
 
       error = function(e) {
         # we're encountering an R error that has not yet been converted to Python
-        trace <- e$trace
-        if(is.null(trace))
-          trace <- get_r_trace(maybe_use_cached = FALSE, trim_tail = 2)
-        e$trace <- .globals$last_r_trace <- trace
+        .globals$last_r_trace <- e$trace <-
+          e$trace %||% get_r_trace(maybe_use_cached = FALSE, trim_tail = 2)
         invokeRestart("raise_py_exception", e)
       }
-    ), # withCallingHandlers()
+    ), # end withCallingHandlers()
 
     raise_py_exception = function(e) {
       list(NULL, e)
     }
-  ) # withRestarts()
+  ) # end withRestarts()
 }
 
 
-as_r_value <- function(x) if(inherits(x, "python.builtin.object")) py_to_r(x) else x
+as_r_value <- function(x)
+  if(inherits(x, "python.builtin.object")) py_to_r(x) else x
 
 
 #' @export
 r_to_py.error <- function(x, convert = FALSE) {
   if(inherits(x, "python.builtin.object")) {
-    assign("convert", convert, envir =  as.environment(x))
+    assign("convert", convert, envir = as.environment(x))
     return(x)
   }
 
-  e <- import_builtins(convert = convert)$RuntimeError(conditionMessage(x))
+  bt <- import_builtins(convert = convert)
+  e <- bt$RuntimeError(conditionMessage(x))
 
-  for (nm in setdiff(names(x), c("call", "message")))
-    py_set_attr(e, paste0("r_", nm), py_capsule(x[[nm]]))
+  for (nm in names(x))
+    py_set_attr(e, nm, py_capsule(x[[nm]]))
 
-  py_set_attr(e, "r_call", conditionCall(x))
-  py_set_attr(e, "r_class", class(x))
+  py_set_attr(e, "r_class", py_capsule(class(x)))
 
   e
 }
 
 #' @export
 conditionCall.python.builtin.BaseException <- function(c) {
-  as_r_value(py_get_attr(c, "r_call", TRUE))
+  as_r_value(py_get_attr(c, "call", TRUE))
 }
 
 #' @export
@@ -177,37 +174,15 @@ conditionMessage.python.builtin.BaseException <- function(c) {
   conditionMessage_from_py_exception(c)
 }
 
-#' @export
-print.python.builtin.BaseException <- function(x, ...) {
-    NextMethod()
-    r_attr_nms <- grep("^r_", py_list_attributes(x), value = TRUE)
-    if (length(r_attr_nms)) {
-      r_attrs <- lapply(r_attr_nms,
-                        function(nm)
-                          as_r_value(py_get_attr(x, nm)))
-      names(r_attrs) <- r_attr_nms
-      r_traceback <- r_attrs$r_traceback
-      r_attrs$r_traceback <- NULL
-      str(r_attrs, no.list = TRUE)
-      if(!is.null(r_traceback)) {
-        cat(" $ r_traceback: \n")
-        traceback(r_traceback)
-      }
-    }
-    invisible(x)
-}
 
 #' @export
 `$.python.builtin.BaseException` <- function(x, name) {
-    if ("condition" %in% .Class &&
-        (identical(name, "call") || identical(name, "message"))) {
-        # warning("Please use conditionCall() or conditionMessage() instead of $call or $message")
-        return(switch(name,
-            call = conditionCall(x),
-            message = conditionMessage(x)
-        ))
-    }
-    py_get_attr(x, name, TRUE)
+  if(identical(name, "call"))
+    return(conditionCall(x))
+  if(identical(name, "message"))
+    return(conditionMessage(x))
+
+  py_maybe_convert(py_get_attr(x, name, TRUE), py_has_convert(x))
 }
 
 #' @export
@@ -416,8 +391,12 @@ canonical_path <- function(path) {
   # on windows we normalize the whole path to avoid
   # short path components leaking in
   if (is_windows()) {
+    # on windows, normalizePath("") returns "C:/"
+    if(isFALSE(nzchar(path))) return("")
     normalizePath(path, winslash = "/", mustWork = FALSE)
   } else {
+    # on linux/mac, we protect against `normalizePath()` resolving
+    # python binaries that are symbolic links, as encountered in python venvs.
     file.path(
       normalizePath(dirname(path), winslash = "/", mustWork = FALSE),
       basename(path)
@@ -537,6 +516,10 @@ isFALSE <- function(x) {
   is.logical(x) && length(x) == 1L && !is.na(x) && !x
 }
 
+is_string <- function(x) {
+  is.character(x) && length(x) == 1L && !is.na(x)
+}
+
 home <- function() {
   path.expand("~")
 }
@@ -599,6 +582,10 @@ str_drop_prefix <- function(x, prefix) {
 
 }
 
+`append<-` <- function(x, value) {
+  c(x, value)
+}
+
 if (getRversion() < "3.3.0") {
 
 startsWith <- function(x, prefix) {
@@ -640,10 +627,17 @@ system2t <- function(command, args, ...) {
   # system2, with a trace
   # mimic bash's set -x usage of a "+" prefix for now
   # maybe someday take a dep on {cli} and make it prettier
-  message(paste("+", shQuote(command), paste0(args, collapse = " ")))
+  message(paste("+", maybe_shQuote(command), paste0(args, collapse = " ")))
   system2(command, args, ...)
 }
 
+maybe_shQuote <- function(x) {
+  needs_quote <- !grepl("^[[:alnum:]/._-]+$", x)
+  if(any(needs_quote))
+    x[needs_quote] <- shQuote(x[needs_quote])
+  x
+}
+
 
 rm_all_reticulate_state <- function() {
   unlink(rappdirs::user_data_dir("r-reticulate", NULL), recursive = TRUE, force = TRUE)
