@@ -41,7 +41,8 @@ config = {
   "BaseExecPrefix"   : getattr(sys, "base_exec_prefix", ""),
   "PythonPath"       : pathsep.join((x or "." for x in sys.path)),
   "LIBPL"            : sysconfig.get_config_var("LIBPL"),
-  "LIBDIR"           : sysconfig.get_config_var("LIBDIR")
+  "LIBDIR"           : sysconfig.get_config_var("LIBDIR"),
+  "SharedLibrary"    : sysconfig.get_config_var("Py_ENABLE_SHARED"),
 }
 
 # detect if this is a conda managed python
